import { locate } from 'locate-character';

var keywords = /\b(case|default|delete|do|else|in|instanceof|new|return|throw|typeof|void)\s*$/;
var punctuators = /(^|\{|\(|\[\.|;|,|<|>|<=|>=|==|!=|===|!==|\+|-|\*\%|<<|>>|>>>|&|\||\^|!|~|&&|\|\||\?|:|=|\+=|-=|\*=|%=|<<=|>>=|>>>=|&=|\|=|\^=|\/=|\/)\s*$/;
var ambiguous = /(\}|\)|\+\+|--)\s*$/;

var punctuatorChars = /[{}()[.;,<>=+\-*%&|\^!~?:/]/;
var keywordChars = /[a-z]/;
var beforeJsxChars = /[=:;,({}[&+]/;

var whitespace = /\s/;

function isWhitespace ( char ) {
	return whitespace.test( char );
}

function isPunctuatorChar ( char ) {
	return punctuatorChars.test( char );
}

function isKeywordChar ( char ) {
	return keywordChars.test( char );
}

function isPunctuator ( str ) {
	return punctuators.test( str );
}

function isKeyword ( str ) {
	return keywords.test( str );
}

function isAmbiguous ( str ) {
	return ambiguous.test( str );
}

function find ( str ) {
	var quote;
	var escapedFrom;
	var regexEnabled = true;
	var pfixOp = false;
	var jsxTagDepth = 0;
	var stack = [];

	var start;
	var found = [];
	var state = base;

	var lsci = -1; // last significant character index
	var lsc = function () { return str[ lsci ]; };

	var parenMatches = {};
	var openingParenPositions = {};
	var parenDepth = 0;

	function tokenClosesExpression () {
		if ( lsc() === ')' ) {
			var c = parenMatches[ lsci ];
			while ( isWhitespace( str[ c - 1 ] ) ) { c -= 1; }

			// if parenthesized expression is immediately preceded by `if`/`while`, it's not closing an expression
			return !/(if|while)$/.test( str.slice( c - 5, c ) );
		}

		// TODO handle }, ++ and -- tokens immediately followed by / character
		return true;
	}

	function base ( char, i ) {
		// the order of these tests is based on which characters are
		// typically more prevalent in a codebase
		if ( char === '(' ) {
			lsci = i;
			openingParenPositions[ parenDepth++ ] = i;
			return base;
		}

		if ( char === ')' ) {
			lsci = i;
			parenMatches[i] = openingParenPositions[ --parenDepth ];
			return base;
		}

		if ( char === '{' ) {
			lsci = i;
			stack.push( base );
			return base;
		}

		if ( char === '}' ) {
			lsci = i;
			return start = i + 1, stack.pop();
		}

		if ( char === '"' || char === "'" ) {
			start = i + 1;
			quote = char;
			stack.push( base );
			return string;
		}

		if ( char === '/' ) {
			// could be start of regex literal OR division punctuator. Solution via
			// http://stackoverflow.com/questions/5519596/when-parsing-javascript-what-determines-the-meaning-of-a-slash/27120110#27120110

			var b = i;
			while ( b > 0 && isWhitespace( str[ b - 1 ] ) ) { b -= 1; }

			if ( b > 0 ) {
				var a = b;

				if ( isPunctuatorChar( str[ a - 1 ] ) ) {
					while ( a > 0 && isPunctuatorChar( str[ a - 1 ] ) ) { a -= 1; }
				} else {
					while ( a > 0 && isKeywordChar( str[ a - 1 ] ) ) { a -= 1; }
				}

				var token = str.slice( a, b );

				regexEnabled = token ? (
					isKeyword( token ) ||
					isPunctuator( token ) ||
					( isAmbiguous( token ) && !tokenClosesExpression() )
				) : false;
			} else {
				regexEnabled = true;
			}

			start = i;
			return slash;
		}

		if ( char === '`' ) {
			start = i + 1;
			return templateString;
		}

		if ( char === '<' && ( !~lsci || beforeJsxChars.test( lsc() ) ) ) {
			stack.push( base );
			return jsxTagStart;
		}

		if ( char === '+' && !pfixOp && str[ i - 1 ] === '+' ) {
			pfixOp = true;
		} else if ( char === '-' && !pfixOp && str[ i - 1 ] === '-' ) {
			pfixOp = true;
		}

		if ( !isWhitespace( char ) ) { lsci = i; }
		return base;
	}

	function slash ( char, i ) {
		if ( char === '/' ) { return start = i + 1, lineComment; }
		if ( char === '*' ) { return start = i + 1, blockComment; }
		if ( char === '[' ) { return regexEnabled ? ( start = i, regexCharacter ) : base; }
		if ( char === '\\' ) { return start = i, escapedFrom = regex, escaped; }
		return regexEnabled && !pfixOp ? ( start = i, regex ) : base;
	}

	function regex ( char, i ) {
		if ( char === '[' ) { return regexCharacter; }
		if ( char === '\\' ) { return escapedFrom = regex, escaped; }

		if ( char === '/' ) {
			var end = i;
			var value = str.slice( start, end );

			found.push({ start: start, end: end, value: value, type: 'regex' });

			return base;
		}

		return regex;
	}

	function regexCharacter ( char ) {
		if ( char === ']' ) { return regex; }
		if ( char === '\\' ) { return escapedFrom = regexCharacter, escaped; }
		return regexCharacter;
	}

	function string ( char, i ) {
		if ( char === '\\' ) { return escapedFrom = string, escaped; }
		if ( char === quote ) {
			var end = i;
			var value = str.slice( start, end );

			found.push({ start: start, end: end, value: value, type: 'string' });

			return stack.pop();
		}

		return string;
	}

	function escaped () {
		return escapedFrom;
	}

	function templateString ( char, i ) {
		if ( char === '$' ) { return templateStringDollar; }
		if ( char === '\\' ) { return escapedFrom = templateString, escaped; }

		if ( char === '`' ) {
			var end = i;
			var value = str.slice( start, end );

			found.push({ start: start, end: end, value: value, type: 'templateEnd' });

			return base;
		}

		return templateString;
	}

	function templateStringDollar ( char, i ) {
		if ( char === '{' ) {
			var end = i - 1;
			var value = str.slice( start, end );

			found.push({ start: start, end: end, value: value, type: 'templateChunk' });

			stack.push( templateString );
			return base;
		}
		return templateString( char, i );
	}

	// JSX is an XML-like extension to ECMAScript
	// https://facebook.github.io/jsx/

	function jsxTagStart ( char ) {
		if ( char === '/' ) { return jsxTagDepth--, jsxTag; }
		return jsxTagDepth++, jsxTag;
	}

	function jsxTag ( char, i ) {
		if ( char === '"' || char === "'" ) { return start = i + 1, quote = char, stack.push( jsxTag ), string; }
		if ( char === '{' ) { return stack.push( jsxTag ), base; }
		if ( char === '>' ) {
			if ( jsxTagDepth <= 0 ) { return base; }
			start = i + 1;
			return jsx;
		}
		if ( char === '/' ) { return jsxTagSelfClosing; }

		return jsxTag;
	}

	function jsxTagSelfClosing ( char ) {
		if ( char === '>' ) {
			jsxTagDepth--;
			if ( jsxTagDepth <= 0 ) { return base; }
			return jsx;
		}

		return jsxTag;
	}

	function jsx ( char, end ) {
		if ( char === '{' || char === '<' ) {
			var value = str.slice( start, end );
			found.push({ start: start, end: end, value: value, type: 'jsx' });

			if ( char === '{' ) {
				return stack.push( jsx ), base;
			}

			if ( char === '<' ) {
				return jsxTagStart;
			}
		}

		return jsx;
	}

	function lineComment ( char, end ) {
		if ( char === '\n' ) {
			var value = str.slice( start, end );

			found.push({ start: start, end: end, value: value, type: 'line' });

			return base;
		}

		return lineComment;
	}

	function blockComment ( char ) {
		if ( char === '*' ) { return blockCommentEnding; }
		return blockComment;
	}

	function blockCommentEnding ( char, i ) {
		if ( char === '/' ) {
			var end = i - 1;
			var value = str.slice( start, end );

			found.push({ start: start, end: end, value: value, type: 'block' });

			return base;
		}

		return blockComment( char );
	}

	for ( var i = 0; i < str.length; i += 1 ) {
		if ( !state ) {
			var ref = locate( str, i, { offsetLine: 1 });
			var line = ref.line;
			var column = ref.column;
			var before = str.slice( 0, i );
			var beforeLine = /(^|\n).+$/.exec( before )[0];
			var after = str.slice( i );
			var afterLine = /.+(\n|$)/.exec( after )[0];

			var snippet = "" + beforeLine + afterLine + "\n" + (Array( beforeLine.length + 1 ).join( ' ' )) + "^";

			throw new Error( ("Unexpected character (" + line + ":" + column + "). If this is valid JavaScript, it's probably a bug in tippex. Please raise an issue at https://github.com/Rich-Harris/tippex/issues – thanks!\n\n" + snippet) );
		}

		state = state( str[i], i );
	}

	// cheeky hack
	if ( state.name === 'lineComment' ) { state( '\n', str.length ); }

	return found;
}

function erase ( str ) {
	var found = find( str );
	return _erase( str, found );
}

function _erase ( str, found ) {
	var erased = '';
	var charIndex = 0;

	for ( var i = 0; i < found.length; i += 1 ) {
		var chunk = found[i];
		erased += str.slice( charIndex, chunk.start );
		erased += chunk.value.replace( /[^\n]/g, ' ' );

		charIndex = chunk.end;
	}

	erased += str.slice( charIndex );
	return erased;
}

function makeGlobalRegExp ( original ) {
	var flags = 'g';

	if ( original.multiline ) { flags += 'm'; }
	if ( original.ignoreCase ) { flags += 'i'; }
	if ( original.sticky ) { flags += 'y'; }
	if ( original.unicode ) { flags += 'u'; }

	return new RegExp( original.source, flags );
}

function match ( str, pattern, callback ) {
	var g = pattern.global;
	if ( !g ) { pattern = makeGlobalRegExp( pattern ); }

	var found = find( str );

	var match;
	var chunkIndex = 0;

	while ( match = pattern.exec( str ) ) {
		var chunk = (void 0);

		do {
			chunk = found[ chunkIndex ];

			if ( chunk && chunk.end < match.index ) {
				chunkIndex += 1;
			} else {
				break;
			}
		} while ( chunk );

		if ( !chunk || chunk.start > match.index ) {
			var args = [].slice.call( match ).concat( match.index, str );
			callback.apply( null, args );
			if ( !g ) { break; }
		}
	}
}

function replace ( str, pattern, callback ) {
	var replacements = [];

	match( str, pattern, function ( match ) {
		var start = arguments[ arguments.length - 2 ];
		var end = start + match.length;
		var content = callback.apply( null, arguments );

		replacements.push({ start: start, end: end, content: content });
	});

	var replaced = '';
	var lastIndex = 0;

	for ( var i = 0; i < replacements.length; i += 1 ) {
		var ref = replacements[i];
		var start = ref.start;
		var end = ref.end;
		var content = ref.content;
		replaced += str.slice( lastIndex, start ) + content;

		lastIndex = end;
	}

	replaced += str.slice( lastIndex );

	return replaced;
}

export { find, erase, match, replace };
//# sourceMappingURL=tippex.es.js.map
